/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2010 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "BSLEARSM_LowRe94w1cp.H"
#include "addToRunTimeSelectionTable.H"
#include "wallFvPatch.H"
#include "backwardsCompatibilityWallFunctions.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace compressible

{
namespace RASModels
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(BSLEARSM_LowRe94w1cp, 0);
addToRunTimeSelectionTable(RASModel, BSLEARSM_LowRe94w1cp, dictionary);

// * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * * //

tmp<volScalarField> BSLEARSM_LowRe94w1cp::F1(const volScalarField& CDkOmega) const
{
    volScalarField CDkOmegaPlus = max
    (
        CDkOmega,
        dimensionedScalar("1.0e-20", dimless/sqr(dimTime), 1.0e-20)
    );
    volScalarField arg1 = min
        (
            max
            (
                (scalar(1)/betaStar_)*sqrt(k_)/(omega_*y_),

                scalar(500)*(mu()/rho_)/(sqr(y_)*omega_)
            ),
            (4*alphaOmega2_)*k_/(CDkOmegaPlus*sqr(y_))
    );

    return tanh(pow4(arg1));
}

    const dimensionedScalar smallOmega_ ("smallOmega",dimensionSet (0,0,-1,0,0,0,0), 1e-9);
    const dimensionedScalar smallDimless_ ("smallDimless",dimless, 1e-9);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

BSLEARSM_LowRe94w1cp::BSLEARSM_LowRe94w1cp
(
    const volScalarField& rho,
    const volVectorField& U,
    const surfaceScalarField& phi,
    const fluidThermo& thermophysicalModel,
    const word& turbulenceModelName,
    const word& modelName
)
:

    RASModel(modelName, rho, U, phi, thermophysicalModel, turbulenceModelName),

    alphaK1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaK1",
            coeffDict_,
            0.5
        )
    ),
    alphaK2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaK2",
            coeffDict_,
            1.0
        )
    ),
    alphaOmega1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaOmega1",
            coeffDict_,
            0.5
        )
    ),
    alphaOmega2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaOmega2",
            coeffDict_,
            0.85616
        )
    ),

    Prt_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Prt",
            coeffDict_,
            1.0
        )
    ),

    gamma1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "gamma1",
            coeffDict_,
            0.5532
        )
    ),
    gamma2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "gamma2",
            coeffDict_,
            0.4403
        )
    ),
    beta1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "beta1",
            coeffDict_,
            0.075
        )
    ),
    beta2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "beta2",
            coeffDict_,
            0.0828
        )
    ),
    betaStar_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "betaStar",
            coeffDict_,
            0.09
        )
    ),
    kappa_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "kappa",
            coeffDict_,
            0.41
        )
    ),
    a1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "a1",
            coeffDict_,
            0.31
        )
    ),
    c1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "c1",
            coeffDict_,
            10.0
        )
    ),

    C1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1",
            coeffDict_,
            1.8
        )
    ),

    A1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "A1",
            coeffDict_,
            1.245
        )
    ),
    y_(mesh_),
    k_
    (
        IOobject
        (
            "k",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateK("k", mesh_)
    ),
    omega_
    (
        IOobject
        (
            "omega",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateOmega("omega", mesh_)
    ),

    tau_( max((1./(betaStar_*omega_)),6.*sqrt((mu()/rho_)/(betaStar_*k_*omega_))) ),

    gradU_( fvc::grad(U_)-(1./3.)*I*tr(fvc::grad(U_)) ),
    S_( tau_*symm(gradU_.T())),
    W_( tau_*skew(gradU_.T())),

    IIs_( tr(S_&S_)),
    IIo_( tr(W_&W_)),
    IV_( tr((S_&W_)&W_)),
    C1prime_( (9./4.)*(C1_-1.)),           //constant
    P1_( C1prime_*((sqr(C1prime_)/27.)+(9.*IIs_/20.)-(2.*IIo_/3.))),
    P2_( sqr(P1_)-pow(((sqr(C1prime_)/9.)+(9.*IIs_/10.)+(2.*IIo_/3.)),3)),
    N_(
        IOobject
        (
            "N",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        C1prime_+(9./4.)*sqrt(2.*betaStar_*IIs_)
    ),
//    N_(scalar(0.)*IIs_),
    Q_( (sqr(N_)-2.*IIo_)/A1_),
    Q1_( Q_*(2.*sqr(N_)-IIo_)/6.),
    Beta1_( -1.*N_/(Q_+smallDimless_)),
    Beta2_( 0.*N_),         //0 work around
    Beta3_( -2.*IV_/(N_*Q1_+smallDimless_)),
    Beta4_( -1./(Q_+smallDimless_)),
    Beta6_( -1.*N_/(Q1_+smallDimless_)),
    Beta9_( 1./(Q1_+smallDimless_)),

    check0_(tr(S_)),        //T1 term
    check1_(tr(((S_ & S_)-(1./3.)*IIs_*I))),        //T2 term
    check2_(tr(((W_ & W_)-(1./3.)*IIo_*I))),        //T3 term
    check3_(tr(((S_ & W_)-(W_ & S_)))),             //T4 term
    check4_(tr((((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_))),   //T6 term
    check5_(tr((((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_))))),  //T9 term

    aij_(
        IOobject
        (
            "aij",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
            symm(
            Beta1_*S_
            //+ Beta2_*((S_ & S_)-(1./3.)*IIs_*I)
            + Beta3_*((W_ & W_)-(1./3.)*IIo_*I)
            + Beta4_*((S_ & W_)-(W_ & S_))
            + Beta6_*(((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_)
            //+ Beta9_*(((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_)))
            )
        ),
    mut_
    (
        IOobject
        (
            "mut",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateMut("mut", mesh_)
    ),

    alphat_
    (
        IOobject
        (
            "alphat",
            runTime_.timeName(),
            mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        autoCreateAlphat("alphat", mesh_)
    ),

    tauij_(
        "tauij",
        symm(
            k_*(aij_+(2./3.)*I)
            )
        ),
    Nij_(
        "Nij",
        symm(

            (k_*aij_+2.0*(mut_/rho_)*symm(gradU_))*rho_

            )
        )
{
    bound(omega_, omegaMin_);

    mut_ = rho_*k_/(omega_+smallOmega_);
    mut_.correctBoundaryConditions();
 

    alphat_ = mut_/Prt_;
    alphat_.correctBoundaryConditions();

    Nij_=symm(k_*aij_+2.0*(mut_/rho_)*symm(gradU_))*rho_;

    k_.correctBoundaryConditions();

    printCoeffs();
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

tmp<volSymmTensorField> BSLEARSM_LowRe94w1cp::R() const
{
    return tmp<volSymmTensorField>
    (
        new volSymmTensorField
        (
            IOobject
            (
                "R",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            ((2.0/3.0)*I)*k_ - (mut_/rho_)*dev(twoSymm(fvc::grad(U_)))+Nij_/rho_,

            //tauij_,
            k_.boundaryField().types()
        )
    );
}


tmp<volSymmTensorField> BSLEARSM_LowRe94w1cp::devRhoReff() const
{
    return tmp<volSymmTensorField>
    (
        new volSymmTensorField
        (
            IOobject
            (
                "devRhoReff",
                runTime_.timeName(),
                mesh_,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),

           -muEff()*dev(twoSymm(fvc::grad(U_))) +dev(Nij_)
        )
    );
}


tmp<fvVectorMatrix> BSLEARSM_LowRe94w1cp::divDevRhoReff(volVectorField& U) const

{
    return
    (
        fvc::div(Nij_)
      - fvm::laplacian(muEff(), U)
      - fvc::div(muEff()*dev2(fvc::grad(U)().T()))
    );
}


bool BSLEARSM_LowRe94w1cp::read()
{
    if (RASModel::read())
    {
        alphaK1_.readIfPresent(coeffDict());
        alphaK2_.readIfPresent(coeffDict());
        alphaOmega1_.readIfPresent(coeffDict());
        alphaOmega2_.readIfPresent(coeffDict());

        Prt_.readIfPresent(coeffDict());

        gamma1_.readIfPresent(coeffDict());
        gamma2_.readIfPresent(coeffDict());
        beta1_.readIfPresent(coeffDict());
        beta2_.readIfPresent(coeffDict());
        betaStar_.readIfPresent(coeffDict());
        a1_.readIfPresent(coeffDict());
        c1_.readIfPresent(coeffDict());

        A1_.readIfPresent(coeffDict());
        C1_.readIfPresent(coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}


void BSLEARSM_LowRe94w1cp::correct()
{

    if (!turbulence_)
    {

        // Re-calculate viscosity
    mut_ = k_*rho_/(omega_+smallOmega_);
        mut_.correctBoundaryConditions();
        // Re-calculate thermal diffusivity
        alphat_ = mut_/Prt_;
        alphat_.correctBoundaryConditions();

        return;
    }

    RASModel::correct();


    /*if(runTime_.timeName() == "1")
    {
      Info<< "SOLVING WALL DISTANCE (Poisson)" << endl;
      tmp<volScalarField> tyPhi_ = tmp<volScalarField>
        (
         new volScalarField
         ( IOobject
           (    "yPhi",
                mesh_.time().timeName(),
                mesh_
            ),
           mesh_,
           dimensionedScalar("yPhi",sqr(dimLength), 1e-15),
           k_.boundaryField().types()
         )
        );

      volScalarField& yPhi = tyPhi_();
      solve(fvm::laplacian(yPhi) == dimensionedScalar("1", dimless, -1.0));
      volVectorField gradyPhi(fvc::grad(yPhi));
      volScalarField magGradyPhi(mag(gradyPhi));
      //y_ = yinit_;
      y_ = sqrt(magSqr(gradyPhi) + 2*yPhi) - magGradyPhi;
      y_ = max(y_,dimensionedScalar("miny",y_.dimensions(),SMALL));
      //y_ = dimensionedScalar("1", dimLength, 1.0);
      y_.write();
    }*/

// NEU:
 volScalarField divU(fvc::div(phi_/fvc::interpolate(rho_)));



    tau_ = max((1./(betaStar_*omega_)),6.*sqrt((mu()/rho_)/(betaStar_*k_*omega_)));

    gradU_ = fvc::grad(U_)-(1./3.)*I*tr(fvc::grad(U_));
    S_ = tau_*symm(gradU_.T());
    W_ = tau_*skew(gradU_.T());

    IIs_ = tr(S_ & S_);
    IIo_ = tr(W_ & W_);
    IV_  = tr((S_ & W_) & W_);
    C1prime_ = (9./4.)*(C1_-1.);           //constant
    P1_ = C1prime_*((sqr(C1prime_)/27.)+(9.*IIs_/20.)-(2.*IIo_/3.));
    P2_ = sqr(P1_)-pow(((sqr(C1prime_)/9.)+(9.*IIs_/10.)+(2.*IIo_/3.)),3);
    forAll(N_,cellI)
    {
        if(P2_[cellI]  >= 0)
        {
        N_[cellI]= (C1prime_.value()/3.)+pow((P1_[cellI]+sqrt(P2_[cellI])),(1./3.))
                    +sign(P1_[cellI]-sqrt(P2_[cellI]))*pow(mag(P1_[cellI]-sqrt(P2_[cellI])),(1./3.));
        }
        else
        {
        N_[cellI]= (C1prime_.value()/3.)+2.*pow((sqr(P1_[cellI])-P2_[cellI]),(1./6.))
                    *cos((1./3.)*acos(P1_[cellI]/(sqrt(sqr(P1_[cellI])-P2_[cellI]))));
        }
    }
    N_.correctBoundaryConditions(); 
    Q_ = (sqr(N_)-2.*IIo_)/A1_;
    Q1_ = Q_*(2.*sqr(N_)-IIo_)/6.;
    Beta1_ = -1.*N_/(Q_+smallDimless_);
    Beta2_ = 0.*N_;     //0 work around
    Beta3_ = -2.*IV_/(N_*Q1_+smallDimless_);
    Beta4_ = -1./(Q_+smallDimless_);
    Beta6_ = -1.*N_/(Q1_+smallDimless_);
    Beta9_ = 1./(Q1_+smallDimless_);

    check0_=tr(S_);                             //T1 term
    check1_=tr(((S_ & S_)-(1./3.)*IIs_*I));        //T2 term
    check2_=tr(((W_ & W_)-(1./3.)*IIo_*I));       //T3 term
    check3_=tr(((S_ & W_)-(W_ & S_)));             //T4 term
    check4_=tr((((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_));   //T6 term
    check5_=tr((((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_))));  //T9 term
//    std::ofstream file("VGlimits.txt");
    Info<< "div(U_)  "<<" max:"<< max(fvc::div(U_)).value() <<
    " min:"<< min(fvc::div(U_)).value() <<" avg:"<< average(fvc::div(U_)).value() << endl;
    Info<< "div(phi_)  "<<" max:"<< max(fvc::div(phi_)).value() <<
    " min:"<< min(fvc::div(phi_)).value() <<" avg:"<< average(fvc::div(phi_)).value() << endl;
    Info<< "gradU_ TRACE  "<<" max:"<< max(tr(gradU_)).value() <<
    " min:"<< min(tr(gradU_)).value() <<" avg:"<< average(tr(gradU_)).value() << endl;
    Info<< "T1 TRACE  "<<" max:"<< max(check0_).value() <<
    " min:"<< min(check0_).value() <<" avg:"<< average(check0_).value() << endl;
    //Info<< "T2 TRACE  "<<" max:"<< max(check1_).value() <<
    //" min:"<< min(check1_).value() <<" avg:"<< average(check1_).value() << endl;
    Info<< "T3 TRACE  "<<" max:"<< max(check2_).value() <<
    " min:"<< min(check2_).value() <<" avg:"<< average(check2_).value() << endl;
    Info<< "T4 TRACE  "<<" max:"<< max(check3_).value() <<
    " min:"<< min(check3_).value() <<" avg:"<< average(check3_).value() << endl;
    Info<< "T6 TRACE  "<<" max:"<< max(check4_).value() <<
    " min:"<< min(check4_).value() <<" avg:"<< average(check4_).value() << endl;
    //Info<< "T9 TRACE  "<<" max:"<< max(check5_).value() <<
    //" min:"<< min(check5_).value() <<" avg:"<< average(check5_).value() << endl;
    Info<< "N        "<<" max:"<< max(N_).value() <<
    " min:"<< min(N_).value() <<" avg:"<< average(N_).value() << endl;
//    file.close();
    /*if(writeResults)
    {
        check1_.write();
        check2_.write();
        check3_.write();
        check4_.write();
        check5_.write();
    }*/

    aij_ = symm(
            Beta1_*S_
            //+ Beta2_*((S_ & S_)-(1./3.)*IIs_*I)
            + Beta3_*((W_ & W_)-(1./3.)*IIo_*I)
            + Beta4_*((S_ & W_)-(W_ & S_))
            + Beta6_*(((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_)
            //+ Beta9_*(((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_)))
            );
    Info<< "aij_ TRACE  "<<" max:"<< max(tr(aij_)).value() <<
    " min:"<< min(tr(aij_)).value() <<" avg:"<< average(tr(aij_)).value() << endl;

    tauij_ = symm(k_*(aij_+(2./3.)*I));
    //Nij_ = symm(tauij_+2.0*nut_*symm(gradU_.T()));       //Non-linear part of Reynolds Stresses
    //Nij_ = symm(k_*aij_+2.0*nut_*symm(gradU_.T()));       //Non-linear part of Reynolds Stresses

    Nij_ = symm(k_*aij_+2.0*(mut_/rho_)*symm(gradU_))*rho_;       //Non-linear part of Reynolds Stresses
// ### Nij_ = symm(k_*aij_+2.0*nut_*symm(gradU_));       //Non-linear part of Reynolds Stresses

//    volScalarField S2 = magSqr(symm(fvc::grad(U_)));
//    volScalarField G("RASModel::G", nut_*2*S2);
//    volScalarField G("RASModel::G", nut_*2*S2- (tauij_&&gradU_) );

    //volScalarField G("RASModel::G", -(tauij_&&gradU_) );
    volScalarField G(GName(), -(tauij_&&gradU_) );

    // Update omega and G at the wall
    omega_.boundaryField().updateCoeffs();
// #		include "kOmega_LowReI.H"

    volScalarField CDkOmega =
        (2*alphaOmega2_)*(fvc::grad(k_) & fvc::grad(omega_))/omega_;            //Cross-diffusion term

    volScalarField F1 = this->F1(CDkOmega);

    // Turbulent frequency equation

/*bisher:
    tmp<fvScalarMatrix> omegaEqn
    (
        fvm::ddt(rho_, omega_)
      + fvm::div(phi_, omega_)
      - fvm::laplacian(DomegaEff(F1), omega_)
     ==
        //fvm::Sp(gamma(F1)*min(G, c1_*betaStar_*k_*omega_)/k_,omega_)        //gamma(F1)*2*S2
        rho_*gamma(F1)*min(G, c1_*betaStar_*k_*omega_)*omega_/k_


      - fvm::Sp(rho_*beta(F1)*omega_, omega_)
      + fvm::Sp
        (
            rho_*(scalar(1)-F1)*CDkOmega/omega_,
            omega_
        )
    );

neu: */

 volScalarField rhoGammaF1(rho_*gamma(F1));

    // Turbulent frequency equation
    tmp<fvScalarMatrix> omegaEqn
    (
        fvm::ddt(rho_, omega_)
      + fvm::div(phi_, omega_)
      - fvm::laplacian(DomegaEff(F1), omega_)
     ==
        rho_*gamma(F1)*min(G, c1_*betaStar_*k_*omega_)*omega_/k_

      - fvm::SuSp((2.0/3.0)*rhoGammaF1*divU, omega_)

      - fvm::Sp(rho_*beta(F1)*omega_, omega_)
      - fvm::SuSp
        (
            rho_*(F1 - scalar(1))*CDkOmega/omega_,
            omega_
        )
);
    omegaEqn().relax();


    omegaEqn().boundaryManipulate(omega_.boundaryField());
// #		include "wallOmega_LowReI.H"

    solve(omegaEqn);
    bound(omega_, omegaMin_);


/*bisher:
    // Turbulent kinetic energy equation
    tmp<fvScalarMatrix> kEqn
    (
        fvm::ddt(rho_, k_)
      + fvm::div(phi_, k_)
      - fvm::laplacian(DkEff(F1), k_)
     ==
        min(G*rho_, c1_*betaStar_*k_*rho_*omega_)
      - fvm::Sp(rho_*betaStar_*omega_, k_)
    );

neu:*/

    // Turbulent kinetic energy equation
    tmp<fvScalarMatrix> kEqn
    (
        fvm::ddt(rho_, k_)
      + fvm::div(phi_, k_)
      - fvm::laplacian(DkEff(F1), k_)
     ==
        min(G*rho_, (c1_*betaStar_)*rho_*k_*omega_)
      - fvm::SuSp(2.0/3.0*rho_*divU, k_) 
      - fvm::Sp(rho_*betaStar_*omega_, k_)
);






    kEqn().relax();
    solve(kEqn);
    bound(k_, kMin_);

    // Re-calculate viscosity

    mut_ = k_*rho_/(omega_+smallOmega_);
    mut_.correctBoundaryConditions();

    tauij_ = symm(k_*(aij_+(2./3.)*I));

    Nij_ = symm(k_*aij_+2.0*(mut_/rho_)*symm(gradU_))*rho_;

    alphat_ = mut_/Prt_;
    alphat_.correctBoundaryConditions();
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace incompressible
} // End namespace Foam

// ************************************************************************* //
